from finesse.cmatrix cimport SubCCSView, SubCCSView1DArray, SubCCSView2DArray
from finesse.knm cimport KnmMatrix
from finesse.cymath cimport complex_t
from finesse.simulations.base cimport ModelData, NodeBeamParam
from finesse.frequency cimport frequency_info_t
from finesse.simulations.basematrix cimport MatrixSystemSolver, CarrierSignalMatrixSimulation
from finesse.element cimport BaseCValues
from finesse.components.workspace cimport ConnectorWorkspace, FillFuncWrapper
from finesse.components.modal.workspace cimport KnmConnectorWorkspace

from finesse.cyexpr cimport (
    cy_expr,
    cy_expr_new,
    cy_expr_init,
    cy_expr_free,
    cy_expr_eval,
)

from cpython.ref cimport PyObject

import numpy as np
cimport numpy as np


cdef struct bs_optical_connections:
    PyObject** P1i_P2o
    PyObject** P2i_P1o
    PyObject** P3i_P4o
    PyObject** P4i_P3o
    PyObject** P1i_P3o
    PyObject** P3i_P1o
    PyObject** P2i_P4o
    PyObject** P4i_P2o


cdef struct bs_signal_connections:
    PyObject*** P1i_Fz
    PyObject*** P1o_Fz
    PyObject*** P2i_Fz
    PyObject*** P2o_Fz
    PyObject*** P3i_Fz
    PyObject*** P3o_Fz
    PyObject*** P4i_Fz
    PyObject*** P4o_Fz
    PyObject*** Z_P1o
    PyObject*** Z_P2o
    PyObject*** Z_P3o
    PyObject*** Z_P4o

    PyObject*** P1i_Fyaw
    PyObject*** P1o_Fyaw
    PyObject*** P2i_Fyaw
    PyObject*** P2o_Fyaw
    PyObject*** P3i_Fyaw
    PyObject*** P3o_Fyaw
    PyObject*** P4i_Fyaw
    PyObject*** P4o_Fyaw
    PyObject*** yaw_P1o
    PyObject*** yaw_P2o
    PyObject*** yaw_P3o
    PyObject*** yaw_P4o

    PyObject*** P1i_Fpitch
    PyObject*** P1o_Fpitch
    PyObject*** P2i_Fpitch
    PyObject*** P2o_Fpitch
    PyObject*** P3i_Fpitch
    PyObject*** P3o_Fpitch
    PyObject*** P4i_Fpitch
    PyObject*** P4o_Fpitch
    PyObject*** pitch_P1o
    PyObject*** pitch_P2o
    PyObject*** pitch_P3o
    PyObject*** pitch_P4o

cdef class BeamsplitterOpticalConnections:
    cdef public:
        Py_ssize_t P1i_P2o_idx
        Py_ssize_t P2i_P1o_idx
        Py_ssize_t P3i_P4o_idx
        Py_ssize_t P4i_P3o_idx
        Py_ssize_t P1i_P3o_idx
        Py_ssize_t P3i_P1o_idx
        Py_ssize_t P2i_P4o_idx
        Py_ssize_t P4i_P2o_idx
    cdef readonly:
        SubCCSView1DArray P1i_P2o
        SubCCSView1DArray P2i_P1o
        SubCCSView1DArray P3i_P4o
        SubCCSView1DArray P4i_P3o
        SubCCSView1DArray P1i_P3o
        SubCCSView1DArray P3i_P1o
        SubCCSView1DArray P2i_P4o
        SubCCSView1DArray P4i_P2o
    cdef:
        bs_optical_connections opt_conn_ptrs


cdef class BeamsplitterSignalConnections(BeamsplitterOpticalConnections):
    cdef public:
        Py_ssize_t P1i_Fz_idx
        Py_ssize_t P1o_Fz_idx
        Py_ssize_t P2i_Fz_idx
        Py_ssize_t P2o_Fz_idx
        Py_ssize_t P3i_Fz_idx
        Py_ssize_t P3o_Fz_idx
        Py_ssize_t P4i_Fz_idx
        Py_ssize_t P4o_Fz_idx
        Py_ssize_t Z_P1o_idx
        Py_ssize_t Z_P2o_idx
        Py_ssize_t Z_P3o_idx
        Py_ssize_t Z_P4o_idx

        Py_ssize_t P1i_Fyaw_idx
        Py_ssize_t P1o_Fyaw_idx
        Py_ssize_t P2i_Fyaw_idx
        Py_ssize_t P2o_Fyaw_idx
        Py_ssize_t P3i_Fyaw_idx
        Py_ssize_t P3o_Fyaw_idx
        Py_ssize_t P4i_Fyaw_idx
        Py_ssize_t P4o_Fyaw_idx
        Py_ssize_t yaw_P1o_idx
        Py_ssize_t yaw_P2o_idx
        Py_ssize_t yaw_P3o_idx
        Py_ssize_t yaw_P4o_idx

        Py_ssize_t P1i_Fpitch_idx
        Py_ssize_t P1o_Fpitch_idx
        Py_ssize_t P2i_Fpitch_idx
        Py_ssize_t P2o_Fpitch_idx
        Py_ssize_t P3i_Fpitch_idx
        Py_ssize_t P3o_Fpitch_idx
        Py_ssize_t P4i_Fpitch_idx
        Py_ssize_t P4o_Fpitch_idx
        Py_ssize_t pitch_P1o_idx
        Py_ssize_t pitch_P2o_idx
        Py_ssize_t pitch_P3o_idx
        Py_ssize_t pitch_P4o_idx
    cdef readonly:
        SubCCSView2DArray P1i_Fz
        SubCCSView2DArray P1o_Fz
        SubCCSView2DArray P2i_Fz
        SubCCSView2DArray P2o_Fz
        SubCCSView2DArray P3i_Fz
        SubCCSView2DArray P3o_Fz
        SubCCSView2DArray P4i_Fz
        SubCCSView2DArray P4o_Fz
        SubCCSView2DArray Z_P1o
        SubCCSView2DArray Z_P2o
        SubCCSView2DArray Z_P3o
        SubCCSView2DArray Z_P4o

        SubCCSView2DArray P1i_Fyaw
        SubCCSView2DArray P1o_Fyaw
        SubCCSView2DArray P2i_Fyaw
        SubCCSView2DArray P2o_Fyaw
        SubCCSView2DArray P3i_Fyaw
        SubCCSView2DArray P3o_Fyaw
        SubCCSView2DArray P4i_Fyaw
        SubCCSView2DArray P4o_Fyaw
        SubCCSView2DArray yaw_P1o
        SubCCSView2DArray yaw_P2o
        SubCCSView2DArray yaw_P3o
        SubCCSView2DArray yaw_P4o

        SubCCSView2DArray P1i_Fpitch
        SubCCSView2DArray P1o_Fpitch
        SubCCSView2DArray P2i_Fpitch
        SubCCSView2DArray P2o_Fpitch
        SubCCSView2DArray P3i_Fpitch
        SubCCSView2DArray P3o_Fpitch
        SubCCSView2DArray P4i_Fpitch
        SubCCSView2DArray P4o_Fpitch
        SubCCSView2DArray pitch_P1o
        SubCCSView2DArray pitch_P2o
        SubCCSView2DArray pitch_P3o
        SubCCSView2DArray pitch_P4o
    cdef:
        bs_signal_connections sig_conn_ptrs


cdef class BeamsplitterValues(BaseCValues):
    cdef public:
        double R
        double T
        double L
        double phi
        double Rcx
        double Rcy
        double xbeta
        double ybeta
        double alpha


cdef class BeamsplitterWorkspace(KnmConnectorWorkspace):
    cdef public:
        complex_t field1_to_F
        complex_t field2_to_F
        complex_t z_to_field1
        complex_t z_to_field2

        # Complete scattering matrices for each propagation direction
        KnmMatrix K12
        KnmMatrix K21
        KnmMatrix K13
        KnmMatrix K31
        KnmMatrix K24
        KnmMatrix K42
        KnmMatrix K34
        KnmMatrix K43

        KnmMatrix K_yaw_sig, K_pitch_sig

        # NOTE (sjr) The above should remain as the full scattering matrices
        #            (i.e. composite over each scattering type) whilst extra
        #            KnmMatrix objects should be stored here for the different
        #            types (i.e. K12_map, K12_aperture, etc.)

        # Arrays of scattering losses for each mode coupling
        double[::1] K12_loss
        double[::1] K21_loss
        double[::1] K13_loss
        double[::1] K31_loss
        double[::1] K24_loss
        double[::1] K42_loss
        double[::1] K34_loss
        double[::1] K43_loss

        double cos_alpha
        double cos_alpha_2

        # Refractive indices of adjacent spaces
        double nr1
        double nr2

        Py_ssize_t car_p1o_rhs_idx, car_p1i_rhs_idx
        Py_ssize_t car_p2o_rhs_idx, car_p2i_rhs_idx
        Py_ssize_t car_p3o_rhs_idx, car_p3i_rhs_idx
        Py_ssize_t car_p4o_rhs_idx, car_p4i_rhs_idx
        Py_ssize_t car_p_num_hom

        # ABCD matrix views
        # -> reflection
        double[:, ::1] abcd_p1p2_x
        double[:, ::1] abcd_p1p2_y
        double[:, ::1] abcd_p2p1_x
        double[:, ::1] abcd_p2p1_y
        double[:, ::1] abcd_p3p4_x
        double[:, ::1] abcd_p3p4_y
        double[:, ::1] abcd_p4p3_x
        double[:, ::1] abcd_p4p3_y
        # -> transmission
        double[:, ::1] abcd_p1p3_x
        double[:, ::1] abcd_p1p3_y
        double[:, ::1] abcd_p3p1_x
        double[:, ::1] abcd_p3p1_y
        double[:, ::1] abcd_p2p4_x
        double[:, ::1] abcd_p2p4_y
        double[:, ::1] abcd_p4p2_x
        double[:, ::1] abcd_p4p2_y

        bint z_signal_enabled
        bint yaw_signal_enabled
        bint pitch_signal_enabled
    cdef:
        BeamsplitterValues cvalues
        BeamsplitterOpticalConnections boc
        BeamsplitterSignalConnections bsc

        # Indices (in sim._c_node_info and sim.trace) of nodes
        Py_ssize_t P1i_id
        Py_ssize_t P1o_id
        Py_ssize_t P2i_id
        Py_ssize_t P2o_id
        Py_ssize_t P3i_id
        Py_ssize_t P3o_id
        Py_ssize_t P4i_id
        Py_ssize_t P4o_id

        # Changing expressions of each ABCD matrix
        # -> size 16 as there are 16 different coupling, plane combos
        # -> order of these corresponds exactly to order of matrix views above
        # NOTE (sjr) Elements A, C & D can change so we will just store arrays
        #            of pointers for every element here, rather than singling out
        #            each element, for convenience
        cy_expr** sym_abcd_elements[16]

        # Direct pointer access to beginning of contiguous memory chunk of
        # each ABCD matrix for speed and convenience in updating
        double* abcd_elements[16]

        frequency_info_t *z_mech_freqs
        Py_ssize_t z_mech_freqs_size
        frequency_info_t *yaw_mech_freqs
        Py_ssize_t yaw_mech_freqs_size
        frequency_info_t *pitch_mech_freqs
        Py_ssize_t pitch_mech_freqs_size

    cpdef compile_abcd_cy_exprs(self)
    cpdef update_parameter_values(self)


cdef object c_beamsplitter_carrier_fill(ConnectorWorkspace cws)
cdef object c_beamsplitter_signal_fill(ConnectorWorkspace cws)



#cython: boundscheck=False, wraparound=False, initializedcheck=False

"""Low-level calculations for Gaussian beam properties.

This module provides functions for computing properties of Gaussian beams
and transforming Gaussian beam parameters via the ABCD ray-matrix framework.

.. note::
    These are low-level functions intended for use in Cython extensions. Users
    should use the functions and classes given in :mod:`finesse.gaussian` instead
    of those defined here.
"""

cpdef complex_t transform_q(
    const double[:, ::1] M,
    complex_t q1,
    double nr1,
    double nr2
) nogil:
    cdef:
        complex_t q1_factor = q1 / nr1

    return nr2 * (M[0][0] * q1_factor + M[0][1]) / (M[1][0] * q1_factor + M[1][1])

cdef complex_t c_transform_q(
    const double* M,
    complex_t q1,
    double nr1,
    double nr2
) nogil:
    """Assumes M is a C-contiguous array of size 2x2"""
    cdef:
        complex_t q1_factor = q1 / nr1

    return nr2 * (M[0] * q1_factor + M[1]) / (M[2] * q1_factor + M[3])

cpdef complex_t inv_transform_q(
    const double[:, ::1] M,
    complex_t q2,
    double nr1,
    double nr2
) nogil:
    cdef:
        complex_t q2_factor = q2 / nr2

    return nr1 * (M[0][1] - M[1][1] * q2_factor) / (M[1][0] * q2_factor - M[0][0])


cdef void abcd_multiply(
    const double[:, ::1] m1,
    const double[:, ::1] m2,
    double[:, ::1] out,
    bint m2_refl_transform=False
) nogil:
    cdef:
        double A1 = m1[0][0]
        double B1 = m1[0][1]
        double C1 = m1[1][0]
        double D1 = m1[1][1]

        double A2 = m2[0][0]
        double B2 = m2[0][1]
        double C2 = m2[1][0]
        double D2 = m2[1][1]

    if m2_refl_transform:
        A2 *= -1.0
        B2 *= -1.0
        C2 *= -1.0
        D2 *= -1.0

    out[0][0] = A1 * A2 + B1 * C2
    out[0][1] = A1 * B2 + B1 * D2
    out[1][0] = C1 * A2 + D1 * C2
    out[1][1] = C1 * B2 + D1 * D2


cdef void sym_abcd_multiply(
    object[:, ::1] m1,
    object[:, ::1] m2,
    object[:, ::1] out,
    bint m2_refl_transform=False
):
    cdef:
        object A1 = m1[0][0]
        object B1 = m1[0][1]
        object C1 = m1[1][0]
        object D1 = m1[1][1]

        object A2 = m2[0][0]
        object B2 = m2[0][1]
        object C2 = m2[1][0]
        object D2 = m2[1][1]

    if m2_refl_transform:
        A2 *= -1.0
        B2 *= -1.0
        C2 *= -1.0
        D2 *= -1.0

    out[0][0] = A1 * A2 + B1 * C2
    out[0][1] = A1 * B2 + B1 * D2
    out[1][0] = C1 * A2 + D1 * C2
    out[1][1] = C1 * B2 + D1 * D2



"""Frequency analysis tools."""

import copy
import weakref
import numpy as np

from finesse.element import ModelElement
from finesse.parameter import float_parameter
from finesse.symbols import Symbol

from libc.stdlib cimport free, malloc

cdef class FrequencyContainer:
    """Contains an array of frequency objects and their associated
    faster access C struct information. 
    """
    def __cinit__(self, *args, **kwargs):
        self.size = 0
        self.frequency_info = NULL
        self.carrier_frequency_info = NULL

    def __init__(self, frequencies, FrequencyContainer carrier_cnt=None):
        self.size = len(frequencies)
        self.frequencies = tuple(frequencies)
        self.frequency_info = <frequency_info_t*> malloc(self.size * sizeof(frequency_info_t))
        if not self.frequency_info:
            raise MemoryError()
        if carrier_cnt:
            if not carrier_cnt.frequency_info:
                raise MemoryError()
            self.carrier_frequency_info = carrier_cnt.frequency_info
        else:
            self.carrier_frequency_info = NULL

    def __dealloc__(self):
        if self.size:
            free(self.frequency_info)

    def get_info(self, Py_ssize_t index):
        if not (0 <= index <= self.size):
            raise IndexError()
        rtn = {
            "f" : self.frequency_info[index].f,
            "index" : self.frequency_info[index].index,
            "audio_lower_index" : self.frequency_info[index].audio_lower_index,
            "audio_upper_index" : self.frequency_info[index].audio_upper_index,
            "audio_order" : self.frequency_info[index].audio_order,
            "audio_carrier_index" : self.frequency_info[index].audio_carrier_index,
        }
        return rtn

    cdef initialise_frequency_info(self):
        cdef Py_ssize_t i, j, cidx
        cdef FrequencyContainer fcnt

        for i in range(self.size):
            self.frequency_info[i].f = <double>self.frequencies[i].f
            self.frequency_info[i].index = <Py_ssize_t>self.frequencies[i].index
            self.frequency_info[i].audio_order = <int>self.frequencies[i].audio_order

            if self.frequency_info[i].audio_order:
                assert(self.carrier_frequency_info)
                cidx = <Py_ssize_t>self.frequencies[i].audio_carrier_index
                self.frequency_info[i].audio_carrier_index = cidx
                self.frequency_info[i].f_car = &self.carrier_frequency_info[cidx].f
                # Update carrier info so it knows about this sideband
                if self.frequency_info[i].audio_order == 1:
                    self.carrier_frequency_info[cidx].audio_upper_index = <Py_ssize_t>self.frequencies[i].index
                elif self.frequency_info[i].audio_order == -1:
                    self.carrier_frequency_info[cidx].audio_lower_index = <Py_ssize_t>self.frequencies[i].index
                else:
                    raise Exception("Unexpected")
        
        # if self.is_audio:
        #     for i in range(len(self.unique_fcnt)):
        #         fcnt = self.unique_fcnt[i]
        #         for j in range(fcnt.size):
        #             fcnt.frequency_info[j].f = <double>fcnt.frequencies[j].f
        #             fcnt.frequency_info[j].index = <Py_ssize_t>fcnt.frequencies[j].index

    cdef update_frequency_info(self):
        cdef Py_ssize_t i, j, cidx
        cdef FrequencyContainer fcnt

        for i in range(self.size):
            self.frequency_info[i].f = <double>self.frequencies[i].f

def generate_frequency_list(model):
    """
    For a given model a symbolic list of frequencies is generated.
    The result can be used to generate a set of frequencies bins to
    be modelled in a simulation.

    This method relies on using :class:`.Symbol`. Using
    symbolic statements this method attempts to isolate uniqe frequency
    bins whilst leaving those changing during a simulation present.

    Returns
    -------
    List of :class:`.Symbol`
    """

    def unique_indices(arr):
        """
        Simple unique element finder which doesn't require any
        greater or less than operations. This isn't tuned for
        efficiency at all.
        """
        lst = list(arr)
        lst2 = list(arr)
        for i in range(len(lst)):
            if lst.count(lst2[i]) > 1:
                lst[i] = None

        return [_ for _, __ in enumerate(lst) if __ is not None]

    fn_eval = np.vectorize(lambda x: x.eval())
    fn_eval2 = np.vectorize(lambda x: x.eval(keep_changing_symbols=True))
    # fn_subs = np.vectorize(
    #     lambda x, **kwargs: x.eval(keep_changing_symbols=True, subs=kwargs)
    # )
    fn_is_changing = np.vectorize(lambda x: x.is_changing)

    source_frequencies = []
    source_components = []
    modulation_frequencies = []

    for comp in model._frequency_generators:
        s = comp._source_frequencies()
        source_frequencies.extend(s)
        source_components.extend((comp,) * len(s))

        m = comp._modulation_frequencies()
        modulation_frequencies.extend(m)

    # Now to prune the frequency list
    Nm = len(modulation_frequencies)
    Ns = len(source_frequencies)

    if Ns == 0:
        raise Exception("There are no source frequencies present in the model")

    if Nm == 0:
        Fsym = np.array(source_frequencies)
    else:
        # First we make a list with all possible combinations of frequencies
        Fsym = (
            np.vstack((np.atleast_2d(modulation_frequencies),) * Ns)
            + np.hstack((np.atleast_2d(source_frequencies).T,) * Nm)
        ).flatten()
        Fsym = np.hstack((source_frequencies, Fsym))

    # Take all the frequency values which definitely won't be changing
    not_changing = Fsym[np.bitwise_not(fn_is_changing(Fsym))]
    if len(not_changing) == 0:
        not_changing = []
    else:
        # ... and select all those which are unique
        values, idx, rev, counts = np.unique(fn_eval(not_changing), True, True, True)
        not_changing = not_changing[idx]

    # First select only the changing frequency bins
    changing = Fsym[fn_is_changing(Fsym)]
    if len(changing) == 0:
        changing = []
    else:
        # We need to use a different unique finding function that doesn't rely on
        # using > or < than comparisons like np.unique. I'm not sure how to implement
        # > and < for Symbols sensibly
        idx = unique_indices(fn_eval2(changing))
        changing = changing[idx]

    # Finally sort the indicies so that upper and lower sidebands are grouped together. This
    # help with numerical errors when iterating outputs later so that similarly sized elements are
    # grouped this isn't always true though, just generally in regards to upper and lower sidebands.
    final = np.hstack((changing, not_changing))
    srt_idx = np.argsort(abs(fn_eval(final)))
    return final[srt_idx]


@float_parameter("f", "Frequency", units="Hz", validate="_validate_fsig", is_default=True)
class Fsig(ModelElement):
    def __init__(self, name, value):
        super().__init__(name)
        self.f = value

    def _validate_fsig(self, value):
        if value is None or isinstance(value, Symbol):
            return value
        elif value <= 0:
            raise Exception("fsig value must be > 0 Hz")
        else:
            return value


cdef class Frequency:
    """Represents a frequency "bin" with a specific index.

    The value of the frequency is calculated from the name of the frequency.

    Parameters
    ----------
    name : str
        Name of the frequency.

    order : int, optional
        The order of the frequency, defaults to zero.
    """
    def __init__(
        self,
        name,
        sim,
        symbol,
        index=None,
        audio=False,
        audio_carrier_index=None,
        audio_order=0,
    ):
        self.__name = name
        self.__symbol = symbol
        self.__index = index
        self.__is_audio = audio
        #self.__sim = sim
        self.__symbol_changing = symbol.is_changing
        self.__start_value = self.__symbol.eval()
        self.__lambdified = eval(f"lambda: {symbol}", sim.model.elements)

        if audio:
            if audio_carrier_index is None:
                raise Exception("Audio frequency carrier must be specified")

            if audio_order not in (-1, 1):
                raise Exception("Audio frequency order must be -1 or +1")

            self.__order = audio_order
            self.__carrier = audio_carrier_index
        else:
            self.__order = 0
            self.__carrier = 0

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} {self.name} (f={self.symbol}={self.f}) at "
            f"{ hex(id(self)) }>"
        )

    def __str__(self):
        return f"<{self.name} is {self.f}Hz (Frequency)>"

    def __deepcopy__(self, memo):
        raise Exception(
            "Frequency objects cannot be deepcopied as they are associated with Simulations"
        )

    #@property
    #def simulation(self):
    #    return self.__sim

    @property
    def symbol(self):
        return self.__symbol

    @property
    def f(self):
        if self.__symbol_changing:
            return self.__lambdified()
        else:
            return self.__start_value

    @property
    def is_audio(self):
        """Is this an audio sideband frequency?"""
        return self.__is_audio

    @property
    def audio_carrier_index(self):
        """The carrier frequency.

        :getter: Returns the carrier frequency index (read-only).
        """
        return self.__carrier

    @property
    def name(self):
        """Name of the frequency object.

        :getter: Returns the name of the frequency (read-only).
        """
        return self.__name

    @property
    def audio_order(self):
        """Audio modulation order of this frequency.

        :getter: Returns the order of the frequency (read-only).
        """
        return self.__order

    @property
    def index(self):
        """Index of the frequency object.

        :getter: Returns the index of the frequency (read-only).
        """
        return self.__index



from finesse.element cimport BaseCValues, ElementWorkspace


cdef class VariableValues(BaseCValues):
    cdef public:
        double value


cdef class VariableWorkspace(ElementWorkspace):
    cdef:
        VariableValues vv



#cython: boundscheck=False, wraparound=False, initializedcheck=False

from finesse.cymath cimport complex_t
from finesse.simulations.base cimport NodeBeamParam
from finesse.cmatrix cimport SubCCSView1DArray

from cpython.ref cimport PyObject

import logging

ctypedef (double*, ) ptr_tuple_1


LOGGER = logging.getLogger(__name__)

# TODO (sjr) make c_isolator_fill function?

cdef class IsolatorValues(BaseCValues):
    def __init__(IsolatorValues self):
        cdef ptr_tuple_1 ptr = (&self.S, )
        cdef tuple params = ("S", )
        self.setup(params, sizeof(ptr), <double**>&ptr)


cdef class IsolatorConnections:
    def __cinit__(self, MatrixSystemSolver mtx):
        size = mtx.optical_frequencies.size
        self.P1i_P2o = SubCCSView1DArray(size)
        self.P2i_P1o = SubCCSView1DArray(size)

        self.conn_ptrs.P1i_P2o = <PyObject**>self.P1i_P2o.views
        self.conn_ptrs.P2i_P1o = <PyObject**>self.P2i_P1o.views


cdef class IsolatorWorkspace(KnmConnectorWorkspace):
    def __init__(self, owner, CarrierSignalMatrixSimulation sim, bint refill):
        super().__init__(
            owner,
            sim,
            refill,
            refill,
            IsolatorConnections(sim.carrier),
            IsolatorConnections(sim.signal),
            IsolatorValues()
        )
        self.P1i_id = sim.carrier.node_id(owner.p1.i)
        self.P1o_id = sim.carrier.node_id(owner.p1.o)
        self.P2i_id = sim.carrier.node_id(owner.p2.i)
        self.P2o_id = sim.carrier.node_id(owner.p2.o)
